/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <pecunia/Money.h>
using pecunia::currency::Money;
#include <QDate>
using ::QDate;
#include <QObject>
using ::QObject;
// using Q_OBJECT
#include <QTest>
// using QTEST_MAIN
// using QVERIFY_EXCEPTION_THROWN
#include <qt-mocks/Matchers.h>
// using QVERIFY_THAT
using qt_mocks::equals;

#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include <banking/BankAccount.h>
using drn::banking::BankAccount;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
using drn::banking::SupportedAccountTypes;
#include <banking/BankingErrors.h>
using drn::banking::BankError;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/Reconciliation.h>
using drn::banking::ReconciledBankAccount;
#include <foundation/Optional.hpp>
using drn::foundation::Optional;
#include <foundation/PresentationText.h>
using ::operator<<;


namespace drn
{
namespace banking
{
namespace tests
{

struct ReconciledBankAccountUT : QObject
{
	Q_OBJECT

	BankAccount bankAccount_{
		BankName{"Home Federal"},
		BankAccountType{
			AccountCode{AccountNumber{1234}, "Personal"},
			SupportedAccountTypes::Chequeing
		}
	};
	Money balance_{21, 43u, Iso4217Codes::USD};
	Optional<QDate> reconciledOn_{QDate::currentDate()};

private slots:
	//{ constructor Tests

	void constructor_default_ShouldBeDefaults()
	{
		BankAccount bankAccount{};
		Money balance{};
		Optional<QDate> reconciledOn{};
		ReconciledBankAccount rba{};
		QVERIFY_THAT(rba.bankAccount(), equals(bankAccount));
		QVERIFY_THAT(rba.balance(), equals(balance));
		QVERIFY_THAT(rba.reconciledOn(), equals(reconciledOn));
	}

	void constructor_partialInit_ShouldBeValues()
	{
		ReconciledBankAccount rba{this->bankAccount_, this->balance_.code()};
		QVERIFY_THAT(rba.bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(rba.balance(), equals(Money{this->balance_.code()}));
		QVERIFY_THAT(rba.reconciledOn(), equals(Optional<QDate>{}));
	}

	void constructor_fullInit_ShouldBeValues()
	{
		ReconciledBankAccount rba{this->bankAccount_, this->balance_, *this->reconciledOn_};
		QVERIFY_THAT(rba.bankAccount(), equals(this->bankAccount_));
		QVERIFY_THAT(rba.balance(), equals(this->balance_));
		QVERIFY_THAT(rba.reconciledOn(), equals(this->reconciledOn_));
	}

	//}

	void reconcile_neverReconciledBefore_ShouldBeNewValues()
	{
		ReconciledBankAccount rba{this->bankAccount_, this->balance_.code()};
		QVERIFY_NO_THROW(rba.reconcile(*this->reconciledOn_, this->balance_));
		QVERIFY_THAT(rba.balance(), equals(this->balance_));
		QVERIFY_THAT(rba.reconciledOn(), equals(this->reconciledOn_));
	}

	void reconcile_reconciledBefore_ShouldBeNewValues()
	{
		ReconciledBankAccount rba{this->bankAccount_, this->balance_.code()};
		QVERIFY_NO_THROW(rba.reconcile(*this->reconciledOn_, this->balance_));
		QVERIFY_NO_THROW(
			rba.reconcile(this->reconciledOn_->addMonths(1), this->balance_ * 2)
		);
		QVERIFY_THAT(rba.balance(), equals(this->balance_ * 2));
		QVERIFY_THAT(
			rba.reconciledOn(),
			equals(Optional<QDate>{this->reconciledOn_->addMonths(1)})
		);
	}

	void reconcile_reconciledOnInThePast_ShouldThrow()
	{
		ReconciledBankAccount rba{this->bankAccount_, this->balance_.code()};
		QVERIFY_NO_THROW(rba.reconcile(*this->reconciledOn_, this->balance_));
		QVERIFY_EXCEPTION_THROWN(
			rba.reconcile(this->reconciledOn_->addMonths(-1), this->balance_),
			BankError
		);
	}
};


}}}

QTEST_MAIN(drn::banking::tests::ReconciledBankAccountUT)
#include "Reconciliation-unit-tests.moc"
